"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.MapAttributeType = void 0;
const lodash_1 = require("lodash");
const errors_1 = require("../../errors");
const attribute_type_1 = require("../../tables/attribute-type");
const truly_empty_1 = require("../../utils/truly-empty");
const util_dynamodb_1 = require("@aws-sdk/util-dynamodb");
class MapAttributeType extends attribute_type_1.AttributeType {
    constructor(record, propertyName, metadata) {
        var _a;
        super(record, propertyName, metadata);
        this.metadata = metadata;
        this.type = "M" /* DynamoAttributeType.Map */;
        this.attributes = {};
        // backwards compatibility to accept 'ignoreUnknownProperties'
        if (this.metadata.arbitraryAttributes == null && this.metadata.ignoreUnknownProperties === true) {
            this.metadata.arbitraryAttributes = 'ignore';
        }
        // convert attributes from ChildAttributeMetadata types to
        for (const childAttributePropertyName of Object.keys(this.metadata.attributes)) {
            const childAttributeDef = (_a = this.metadata) === null || _a === void 0 ? void 0 : _a.attributes[childAttributePropertyName];
            const childAttribute = childAttributeDef.getAttribute(record, childAttributePropertyName);
            this.attributes[childAttribute.name] = childAttribute;
        }
    }
    getDefault() {
        const map = {};
        for (const childAttribute of Object.values(this.attributes)) {
            map[childAttribute.name] = childAttribute.getDefaultValue();
        }
        return map;
    }
    toDynamo(mapValue) {
        var _a;
        if (!(0, lodash_1.isObject)(mapValue)) {
            throw new errors_1.ValidationError(`Map attributes require values to be an Object, but was given a ${typeof mapValue}`);
        }
        const map = {};
        const marshallOptions = (_a = this.metadata.marshallOptions) !== null && _a !== void 0 ? _a : { convertEmptyValues: true, removeUndefinedValues: true };
        for (const propertyName of Object.keys(mapValue)) {
            const attribute = (0, lodash_1.find)(this.attributes, (attr) => attr.propertyName === propertyName);
            const value = (0, lodash_1.get)(mapValue, propertyName);
            if (attribute != null) {
                const attributeValue = attribute.toDynamo(value);
                if (attributeValue != null) {
                    map[attribute.name] = attributeValue;
                }
            }
            else if (this.metadata.arbitraryAttributes === 'marshall') {
                if (value != null) {
                    map[propertyName] = (0, util_dynamodb_1.marshall)({ value }, marshallOptions).value;
                }
            }
            else if (this.metadata.arbitraryAttributes !== 'ignore') {
                throw new errors_1.ValidationError(`Unknown property set on Map, ${propertyName}`);
            }
        }
        return { M: map };
    }
    fromDynamo(attributeValue) {
        const mapValue = attributeValue.M == null ? {} : attributeValue.M;
        const map = {};
        for (const attributeName of Object.keys(mapValue)) {
            const value = mapValue[attributeName];
            const attribute = this.attributes[attributeName];
            if (attribute != null) {
                map[attribute.propertyName] = attribute.fromDynamo(value);
            }
            else if (this.metadata.arbitraryAttributes === 'marshall') {
                map[attributeName] = (0, util_dynamodb_1.unmarshall)({ value }, this.metadata.unmarshallOptions).value;
            }
            else if (this.metadata.arbitraryAttributes !== 'ignore') {
                throw new errors_1.ValidationError(`Unknown attribute seen on Map, ${attributeName}`);
            }
        }
        return map;
    }
    fromJSON(json) {
        const mapValue = {};
        (0, lodash_1.each)(json, (value, propertyName) => {
            const attribute = (0, lodash_1.find)(this.attributes, (attr) => attr.propertyName === propertyName);
            if (attribute != null) {
                if (typeof attribute.type.fromJSON === 'function') {
                    value = attribute.type.fromJSON(value);
                }
                // compare to current value, to avoid unnecessarily marking attributes as needing to be saved
                mapValue[propertyName] = value;
            }
        });
        return mapValue;
    }
    toJSON(mapValue) {
        const json = {};
        for (const propertyName of Object.keys(mapValue)) {
            const attribute = (0, lodash_1.find)(this.attributes, (attr) => attr.propertyName === propertyName);
            const value = (0, lodash_1.get)(mapValue, propertyName);
            if (attribute != null) {
                if (!(0, truly_empty_1.isTrulyEmpty)(value)) {
                    if ((0, lodash_1.isFunction)(attribute.type.toJSON)) {
                        json[propertyName] = attribute.type.toJSON(value, attribute);
                    }
                    else {
                        json[propertyName] = value;
                    }
                }
            }
            else if (this.metadata.arbitraryAttributes === 'marshall') {
                json[propertyName] = value;
            }
            else if (this.metadata.arbitraryAttributes !== 'ignore') {
                throw new errors_1.ValidationError(`Unknown property set on Map during toJSON, ${propertyName}`);
            }
        }
        return json;
    }
}
exports.MapAttributeType = MapAttributeType;
//# sourceMappingURL=map.js.map